@@ -92,6 +92,7 @@
 #include "xdriver-shm-cmd.h"
 #include "qubes.h"
 #include "labels.h"
+#include "unix-addr.h"
 
 #include "../../xf86-qubes-common/include/xf86-qubes-common.h"
 
@@ -329,8 +330,8 @@ static void QubesInitAxesLabels(QubesDevicePtr pQubes, int natoms,
     const char **labels;
     int labels_len = 0;
 
-    labels = rel_labels;
-    labels_len = ArrayLength(rel_labels);
+    labels = abs_labels;
+    labels_len = ArrayLength(abs_labels);
 
     memset(atoms, 0, natoms * sizeof(Atom));
 
@@ -361,11 +362,12 @@ static int _qubes_init_axes(DeviceIntPtr device)
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                                        atoms,
 #endif
-                                       GetMotionHistorySize(), 0))
+                                       GetMotionHistorySize(),
+                                       Absolute))
         return BadAlloc;
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
-    pInfo->dev->valuator->mode = Relative;
+    pInfo->dev->valuator->mode = Absolute;
 #endif
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 13
     if (!InitAbsoluteClassDeviceStruct(device))
@@ -373,10 +375,10 @@ static int _qubes_init_axes(DeviceIntPtr device)
 #endif
 
     for (i = 0; i < pQubes->axes; i++) {
-        xf86InitValuatorAxisStruct(device, i, *pQubes->labels, -1,
-                -1, 1, 1, 1
+        xf86InitValuatorAxisStruct(device, i, atoms[i], 0,
+                32767, 1, 1, 1
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-                , Relative
+                , Absolute
 #endif
                 );
         xf86InitValuatorDefaults(device, i);
@@ -387,19 +389,22 @@ static int _qubes_init_axes(DeviceIntPtr device)
 int connect_unix_socket(QubesDevicePtr pQubes);
 int connect_unix_socket(QubesDevicePtr pQubes)
 {
-    int s, len;
+    int s;
     struct sockaddr_un remote;
+    socklen_t len;
+
+    len = sockaddr_un_from_path(&remote, pQubes->device);
+    if (len == 0) {
+        xf86Msg(X_ERROR, "invalid socket path: %s\n", pQubes->device);
+        return -1;
+    }
 
     if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
         xf86Msg(X_ERROR, "socket(%s): %s\n", pQubes->device, strerror(errno));
         return -1;
     }
 
-
-    remote.sun_family = AF_UNIX;
-    strncpy(remote.sun_path, pQubes->device, sizeof(remote.sun_path));
-    len = strlen(remote.sun_path) + sizeof(remote.sun_family);
-    if (connect(s, (struct sockaddr *) &remote, len) == -1) {
+    if (connect(s, (struct sockaddr *)&remote, len) == -1) {
         xf86Msg(X_ERROR, "connect(%s): %s\n", pQubes->device, strerror(errno));
         close(s);
         return -1;
@@ -599,6 +604,19 @@ static void dump_window_grant_refs(int window_id, int fd)
         // the window is destroyed before the driver sees the req
         goto send_response;
 
+    if (!x_window->realized) {
+        // Composite redirect is setup/teared-down during Realize-/Unrealize-
+        // Window (called when a window gets mapped/unmapped). So if the window
+        // is not realized we don't have a per window pixmap we can send.
+        //
+        // This means that the configure of a window before mapping will not
+        // send grants. This should be fine since when compRealizeWindow
+        // replaces the pixmap it will trigger a damage event and we will send
+        // the grants then.
+        xf86Msg(X_ERROR, "can't dump not realized window\n");
+        goto send_response;
+    }
+
     screen = x_window->drawable.pScreen;
     pixmap = (*screen->GetWindowPixmap) (x_window);
 
