@@ -48,10 +48,6 @@
 #include <X11/Xproto.h>
 #include "scrnintstr.h"
 #include "servermd.h"
-#ifdef USE_DGA
-#define _XF86DGA_SERVER_
-#include <X11/extensions/xf86dgaproto.h>
-#endif
 
 /* glamor support */
 #define GLAMOR_FOR_XORG
@@ -87,6 +83,16 @@ static Bool     dummyDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op,
 /* static void     DUMMYDisplayPowerManagementSet(ScrnInfoPtr pScrn, */
 /*          int PowerManagementMode, int flags); */
 
+/*
+ * Make FBBase grant-table backed if glamor is initialized. Currently it's
+ * believed to be unneeded, but if you get the
+ * "can't dump window without grant table allocation" message for a root
+ * window, set DUMMY_GLAMOR_GNT_BACKED_FBBASE to 1 and report an issue.
+ * See discussion at:
+ * https://github.com/QubesOS/qubes-gui-agent-linux/pull/233
+ */
+#define DUMMY_GLAMOR_GNT_BACKED_FBBASE 0
+
 #define DUMMY_VERSION 4000
 #define DUMMY_NAME "DUMMYQBS"
 #define DUMMY_DRIVER_NAME "dummyqbs"
@@ -131,8 +137,8 @@ _X_EXPORT DriverRec DUMMY = {
 };
 
 static SymTabRec DUMMYChipsets[] = {
-    { DUMMY_CHIP,   "dummy" },
-    { -1,		 NULL }
+    { DUMMY_CHIP, "dummy" },
+    { -1,         NULL }
 };
 
 typedef enum {
@@ -142,10 +148,10 @@ typedef enum {
 } DUMMYOpts;
 
 static const OptionInfoRec DUMMYOptions[] = {
-    { OPTION_SW_CURSOR,	"SWcursor",	OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_RENDER,	"Render",	OPTV_STRING,	{0}, FALSE },
+    { OPTION_SW_CURSOR, "SWcursor",     OPTV_BOOLEAN,   {0}, FALSE },
+    { OPTION_RENDER,    "Render",       OPTV_STRING,    {0}, FALSE },
     { OPTION_GUI_DOMID, "GUIDomID",     OPTV_INTEGER,   {0}, FALSE },
-    { -1,                  NULL,           OPTV_NONE,	{0}, FALSE }
+    { -1,                  NULL,           OPTV_NONE,   {0}, FALSE }
 };
 
 #ifdef XFree86LOADER
@@ -410,13 +416,41 @@ Bool DUMMYAdjustScreenPixmap(ScrnInfoPtr pScrn, int width, int height)
                 "Failed to get the screen pixmap.\n");
         return FALSE;
     }
-    if (cbLine > UINT32_MAX || cbLine * height > pScrn->videoRam * 1024)
-    {
+    if (cbLine > UINT32_MAX) {
         xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
-                "Unable to set up a virtual screen size of %dx%d with %d Kb of video memory available.  Please increase the video memory size.\n",
-                width, height, pScrn->videoRam);
+                "Unable to set up a virtual screen size of %dx%d, cbLine "
+                "overflow\n",
+                width, height);
         return FALSE;
     }
+    if (cbLine * height > pScrn->videoRam * 1024) {
+        if (!dPtr->FBBasePriv) {
+            /* If there is no backing grant entries, it's easy enough to extend
+             */
+            pointer *newFBBase;
+            size_t new_size = (cbLine * height + 1023) & ~1023;
+
+            newFBBase = realloc(dPtr->FBBase, new_size);
+            if (!newFBBase) {
+                xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+                        "Unable to set up a virtual screen size of %dx%d, "
+                        "cannot allocate memory (%zu bytes)\n",
+                        width, height, new_size);
+                return FALSE;
+            }
+            memset((char*)newFBBase + pScrn->videoRam * 1024,
+                   0,
+                   new_size - pScrn->videoRam * 1024);
+            dPtr->FBBase = newFBBase;
+            pScrn->videoRam = new_size / 1024;
+        } else {
+            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+                    "Unable to set up a virtual screen size of %dx%d with %d Kb of video memory available.  Please increase the video memory size.\n",
+                    width, height, pScrn->videoRam);
+            return FALSE;
+        }
+    }
+
     pScreen->ModifyPixmapHeader(pPixmap, width, height,
             pScrn->depth, xf86GetBppFromDepth(pScrn, pScrn->depth), cbLine,
             dPtr->FBBase);
@@ -534,7 +568,7 @@ DUMMYProbe(DriverPtr drv, int flags)
 
         for (i = 0; i < numUsed; i++) {
             ScrnInfoPtr pScrn = NULL;
-            int entityIndex = 
+            int entityIndex =
                 xf86ClaimNoSlot(drv,DUMMY_CHIP,devSections[i],TRUE);
             /* Allocate a ScrnInfoRec and claim the slot */
             if ((pScrn = xf86AllocateScreen(drv,0 ))) {
@@ -555,7 +589,7 @@ DUMMYProbe(DriverPtr drv, int flags)
                 foundScreen = TRUE;
             }
         }
-    }    
+    }
     return foundScreen;
 }
 
@@ -575,21 +609,21 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
     GDevPtr device = xf86GetEntityInfo(pScrn->entityList[0])->device;
     const char *render, *defaultRender = "/dev/dri/renderD128";
 
-    if (flags & PROBE_DETECT) 
+    if (flags & PROBE_DETECT)
         return TRUE;
-    
+
     /* Allocate the DummyRec driverPrivate */
     if (!DUMMYGetRec(pScrn)) {
         return FALSE;
     }
-    
+
     dPtr = DUMMYPTR(pScrn);
 
     pScrn->chipset = (char *)xf86TokenToString(DUMMYChipsets,
             DUMMY_CHIP);
 
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Chipset is a DUMMY\n");
-    
+
     pScrn->monitor = pScrn->confScreen->monitor;
 
     if (!xf86SetDepthBpp(pScrn, 0, 0, 0,  Support24bppFb | Support32bppFb))
@@ -634,7 +668,7 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
         }
     }
 
-    if (!xf86SetDefaultVisual(pScrn, -1)) 
+    if (!xf86SetDefaultVisual(pScrn, -1))
         return FALSE;
 
     if (pScrn->depth > 1) {
@@ -684,8 +718,8 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
     clockRanges->ClockMulFactor = 1;
     clockRanges->minClock = 11000;   /* guessed §§§ */
     clockRanges->maxClock = maxClock;
-    clockRanges->clockIndex = -1;		/* programmable */
-    clockRanges->interlaceAllowed = TRUE; 
+    clockRanges->clockIndex = -1; /* programmable */
+    clockRanges->interlaceAllowed = TRUE;
     clockRanges->doubleScanAllowed = TRUE;
 
     /* Subtract memory for HW cursor */
@@ -721,8 +755,8 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
      * driver and if the driver doesn't provide code to set them.  They
      * are not pre-initialised at all.
      */
-    xf86SetCrtcForModes(pScrn, 0); 
- 
+    xf86SetCrtcForModes(pScrn, 0);
+
     /* Set the current mode to the first in the list */
     pScrn->currentMode = pScrn->modes;
 
@@ -740,7 +774,7 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
         if (!xf86LoadSubModule(pScrn, "ramdac"))
             RETURN;
     }
-    
+
     /* We have no contiguous physical fb in physical memory */
     pScrn->memPhysBase = 0;
     pScrn->fbOffset = 0;
@@ -750,7 +784,7 @@ DUMMYPreInit(ScrnInfoPtr pScrn, int flags)
 
     if (!render)
         render = defaultRender;
- 
+
     dPtr->fd = open(render, O_RDWR);
     if (dPtr->fd < 0)
         xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Open render %s fail\n", render);
@@ -775,7 +809,7 @@ static Bool
 DUMMYEnterVT(VT_FUNC_ARGS_DECL)
 {
     SCRN_INFO_PTR(arg);
-    
+
     /* Should we re-save the text mode on each VT enter? */
     if(!dummyModeInit(pScrn, pScrn->currentMode))
       return FALSE;
@@ -809,7 +843,7 @@ DUMMYLoadPalette(
        shift = Gshift = 1;
        break;
    case 16:
-       shift = 0; 
+       shift = 0;
        Gshift = 0;
        break;
    default:
@@ -822,7 +856,7 @@ DUMMYLoadPalette(
        dPtr->colors[index].red = colors[index].red << shift;
        dPtr->colors[index].green = colors[index].green << Gshift;
        dPtr->colors[index].blue = colors[index].blue << shift;
-   } 
+   }
 
 }
 
@@ -915,9 +949,10 @@ qubes_create_screen_resources(ScreenPtr pScreen) {
     if (ret) {
         PixmapPtr pixmap = pScreen->GetScreenPixmap(pScreen);
         if (dPtr->glamor)
-           glamor_egl_create_textured_pixmap_from_gbm_bo(pixmap, dPtr->front_bo, FALSE);
-        xf86_qubes_pixmap_set_private(pixmap,
-                                      dPtr->FBBasePriv);
+            glamor_egl_create_textured_pixmap_from_gbm_bo(pixmap, dPtr->front_bo, FALSE);
+        if (dPtr->FBBasePriv)
+            xf86_qubes_pixmap_set_private(pixmap,
+                                          dPtr->FBBasePriv);
     }
 
     return ret;
@@ -1015,16 +1050,23 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
         return FALSE;
     }
 
-    dPtr->FBBasePriv = qubes_alloc_pixmap_private(pScrn->videoRam * 1024);
-    if (dPtr->FBBasePriv == NULL)
-        return FALSE;
-    dPtr->FBBase = (void *) dPtr->FBBasePriv->data;
-    
+    if (DUMMY_GLAMOR_GNT_BACKED_FBBASE && dPtr->glamor) {
+        dPtr->FBBasePriv = qubes_alloc_pixmap_private(pScrn->videoRam * 1024);
+        if (dPtr->FBBasePriv == NULL)
+            return FALSE;
+        dPtr->FBBase = (void *) dPtr->FBBasePriv->data;
+    } else {
+        dPtr->FBBase = calloc(1, pScrn->videoRam * 1024);
+        if (dPtr->FBBase == NULL)
+            return FALSE;
+    }
+
+
     /*
      * next we save the current state and setup the first mode
      */
     dummySave(pScrn);
-    
+
     if (!dummyModeInit(pScrn,pScrn->currentMode))
         return FALSE;
     DUMMYAdjustFrame(ADJUST_FRAME_ARGS(pScrn, pScrn->frameX0, pScrn->frameY0));
@@ -1033,9 +1075,9 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
      * Reset visual list.
      */
     miClearVisualTypes();
-    
+
     /* Setup the visuals we support. */
-    
+
     if (!miSetVisualTypes(pScrn->depth,
                 miGetDefaultVisualMask(pScrn->depth),
                 pScrn->rgbBits, pScrn->defaultVisual))
@@ -1119,10 +1161,6 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
     dPtr->CreateScreenResources = pScreen->CreateScreenResources;
     pScreen->CreateScreenResources = qubes_create_screen_resources;
 
-#ifdef USE_DGA
-    DUMMYDGAInit(pScreen);
-#endif
-
     /* initialize XRANDR */
     xf86CrtcConfigInit(pScrn, &DUMMYCrtcConfigFuncs);
     /* FIXME */
@@ -1173,7 +1211,7 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
     }
 
     /* XRANDR initialization end */
-    
+
     if (dPtr->swCursor)
         xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Using Software Cursor.\n");
 
@@ -1187,9 +1225,9 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
         AvailFBArea.y1 = 0;
         AvailFBArea.x2 = pScrn->displayWidth;
         AvailFBArea.y2 = lines;
-        xf86InitFBManager(pScreen, &AvailFBArea); 
+        xf86InitFBManager(pScreen, &AvailFBArea);
 
-        xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                 "Using %i scanlines of offscreen memory \n"
                 , lines - pScrn->virtualY);
     }
@@ -1215,8 +1253,8 @@ DUMMYScreenInit(SCREEN_INIT_ARGS_DECL)
         return FALSE;
 
     if (!xf86HandleColormaps(pScreen, 256, pScrn->rgbBits,
-                DUMMYLoadPalette, NULL, 
-                CMAP_PALETTED_TRUECOLOR 
+                DUMMYLoadPalette, NULL,
+                CMAP_PALETTED_TRUECOLOR
                 | CMAP_RELOAD_ON_MODE_SWITCH))
         return FALSE;
 
@@ -1254,7 +1292,7 @@ DUMMYSwitchMode(SWITCH_MODE_ARGS_DECL)
 DUMMYAdjustFrame(ADJUST_FRAME_ARGS_DECL)
 {
     SCRN_INFO_PTR(arg);
-    int Base; 
+    int Base;
 
     Base = (y * pScrn->displayWidth + x) >> 2;
 
@@ -1290,8 +1328,12 @@ DUMMYCloseScreen(CLOSE_SCREEN_ARGS_DECL)
         if (dPtr->FBBasePriv) {
             xf86_qubes_free_pixmap_private(dPtr->FBBasePriv);
             dPtr->FBBasePriv = NULL;
-            dPtr->FBBase = NULL;
+        } else {
+            if (dPtr->FBBase) {
+                free(dPtr->FBBase);
+            }
         }
+        dPtr->FBBase = NULL;
     }
 
     if (dPtr->CursorInfo)
@@ -1321,7 +1363,7 @@ DUMMYSaveScreen(ScreenPtr pScreen, int mode)
         dPtr = DUMMYPTR(pScrn);
 
         dPtr->screenSaver = xf86IsUnblank(mode);
-    } 
+    }
     return TRUE;
 }
 
@@ -1337,7 +1379,7 @@ dummySave(ScrnInfoPtr pScrn)
 {
 }
 
-    static void 
+    static void
 dummyRestore(ScrnInfoPtr pScrn, Bool restoreText)
 {
 }
@@ -1379,7 +1421,7 @@ DUMMYCreateWindow(WindowPtr pWin)
             VFB_PROP = MakeAtom(VFB_PROP_NAME, strlen(VFB_PROP_NAME), 1);
 
 #if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 21
-        ret = ChangeWindowProperty(pWinRoot, VFB_PROP, XA_STRING, 
+        ret = ChangeWindowProperty(pWinRoot, VFB_PROP, XA_STRING,
                 8, PropModeReplace, (int)4, (pointer)"TRUE", FALSE);
 #else
         ret = dixChangeWindowProperty(serverClient, pWinRoot,
